/***********************************************************
Common software for servo function.
Author : 
create : 01/10/2013 version 1.0
modify :
***********************************************************/
//----------------------------------------------------------
// Includes
//----------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <cv.h>
#include <highgui.h> 

//----------------------------------------------------------
// Defines
//----------------------------------------------------------
#define	SIZE_BUFFER	256
#define	VALUE_THRESHHOLD	40
#define	MINI_AREA	50

#define	PR(A,B)	printf("--->%s.(%s)\n", A, B)
#define PRD(A,B) printf("DEBUG--->%s.(%s)\n", A, B)
//----------------------------------------------------------
// Globals
//----------------------------------------------------------
//----------------------------------------------------------
// Prototype
//----------------------------------------------------------
void cv_HsvExtraction(IplImage* ipSrcImg, IplImage* ipResultImg,
		int iCh1Lower, int iCh1Upper,int iCh2Lower, int iCh2Upper,
		int iCh3Lower, int iCh3Upper);
//----------------------------------------------------------
// Code
//----------------------------------------------------------

void cv_HsvExtraction(
	IplImage* ipSrcImg,		// 元画像
	IplImage* ipResultImg,  // 抽出画像
	int iCh1Lower,		// HUE下限値
	int iCh1Upper,		// HUE上限値
	int iCh2Lower,		// SATURATION下限値
	int iCh2Upper,		// SATURATION上限値
	int iCh3Lower, 		// VALUE下限値
	int iCh3Upper		// VALUE上限値
	){

	int i, j;
	int iLower[3];
	int iUpper[3];
	int iScValue[3];

	IplImage* ipHsvImg;
	IplImage* ipCvtImg;
	IplImage* ipCh1Img;
	IplImage* ipCh2Img;
	IplImage* ipCh3Img;
	IplImage* ipMskImg;

	CvMat* Lut;
    CvScalar Sc;

	/*******************************************************
	** Initialize 
	*******************************************************/
	/*******************************************************
	** Process 
	*******************************************************/
	// BRGからHSV色空間に変換
	ipHsvImg = cvCreateImage(cvGetSize(ipSrcImg), ipSrcImg->depth, ipSrcImg->nChannels);
	cvCvtColor(ipSrcImg, ipHsvImg, CV_BGR2HSV);

	//------------------------------------------------------
	// マスクデータ作成のために
	// HSV成分毎の範囲指定値を用いて
	// 二値化したLook-up table（出力値の集合）を作成する
	// LUT変換によりマスク用データを生成する
	
	Lut	= cvCreateMat(256, 1, CV_8UC3); // 符号無し8bit3CH行列生成

	// HSV3成分毎の上限、下限値判定値の設定
	iLower[0] = iCh1Lower;
	iLower[1] = iCh2Lower;
	iLower[2] = iCh3Lower;

	iUpper[0] = iCh1Upper;
	iUpper[1] = iCh2Upper;
	iUpper[2] = iCh3Upper;

	// 成分毎の範囲値から、成分毎の二値化した変換用テーブル作成
	for (i = 0; i < 256; i++){
		for (j = 0; j < 3; j++){
			if (iLower[j] <= iUpper[j]){
				if ((iLower[j] <= i) && (i <= iUpper[j])){
					iScValue[j] = 255;
				}else{
					iScValue[j] = 0;
				}
			}else{	// 0°, 0%を挟んだ設定対応処理
				if ((i <= iUpper[j]) || (iLower[j] <= i)){
					iScValue[j] = 255;
				}else{
					iScValue[j] = 0;
				}
			}
		}
		// 出力用の値の設定
		Sc = cvScalar(iScValue[0], iScValue[1], iScValue[2], 0.0);
		cvSet1D(Lut, i, Sc);
	}

	// LUT変換
	ipCvtImg = cvCreateImage(cvGetSize(ipSrcImg), ipSrcImg->depth, ipSrcImg->nChannels);
	cvLUT(ipHsvImg, ipCvtImg, Lut);

	//------------------------------------------------------
	// マスク用データを各チャンネルに分割して、
	// 各チャンネルのAND処理を行い、マスクを作成する。
	// マスクを使い、抽出データを生成する。（HSVを使った逆変換）
	
	// 各チャンネル毎にイメージ領域を確保
	ipCh1Img = cvCreateImage(cvGetSize(ipCvtImg), ipCvtImg->depth, 1);
	ipCh2Img = cvCreateImage(cvGetSize(ipCvtImg), ipCvtImg->depth, 1);
	ipCh3Img = cvCreateImage(cvGetSize(ipCvtImg), ipCvtImg->depth, 1);

	// マルチチャンネルをシングルチャンネル分解
	cvSplit(ipCvtImg, ipCh1Img, ipCh2Img, ipCh3Img, NULL);

	// マスク画像合成
	ipMskImg = cvCreateImage(cvGetSize(ipCvtImg), ipCvtImg->depth, 1);
	cvAnd(ipCh1Img, ipCh2Img, ipMskImg, NULL);
	cvAnd(ipMskImg, ipCh3Img, ipMskImg, NULL);

	// マスク画像を使って、出力画像を生成する
	cvSetZero(ipResultImg);
	cvCopy(ipSrcImg, ipResultImg, ipMskImg);

	/*******************************************************
	** Terminate 
	*******************************************************/
	// Release
	cvReleaseImage(&ipHsvImg);
	cvReleaseMat(&Lut);
	cvReleaseImage(&ipCvtImg);
	cvReleaseImage(&ipCh1Img);
	cvReleaseImage(&ipCh2Img);
	cvReleaseImage(&ipCh3Img);
	cvReleaseImage(&ipMskImg);

	return;
}