/***********************************************************
Common software for servo function.
Author : 
create : 01/10/2013 version 1.0
modify :
***********************************************************/
//----------------------------------------------------------
// Includes
//----------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
//opencv
#include <cv.h>
#include <highgui.h>

//----------------------------------------------------------
// Defines
//----------------------------------------------------------
#define MAX_GRAVITY	512

#define	PR(A,B)	printf("--->%s.(%s)\n", A, B)
#define PRD(A,B) printf("DEBUG--->%s.(%d)\n", A, B)

//----------------------------------------------------------
// Globals
//----------------------------------------------------------
int	iCntArea = 0;
int	iGravityX[MAX_GRAVITY];
int iGravityY[MAX_GRAVITY];
int iArea[MAX_GRAVITY];          // area
int	iAreaCount[MAX_GRAVITY];     // number of coordinates of area

//----------------------------------------------------------
// Prototype
//----------------------------------------------------------
int  cv_Labelling(IplImage *ipSrcImg, IplImage *ipDstImg, int iThresh);
void rx_DrawArea (IplImage *ipWkImg, CvSeq *stContour, int iLevel,	int	iThresh);
void rx_DrawAreaChild(IplImage *ipWkImg, CvSeq *stContour, int iLevel, int	iThresh);

//----------------------------------------------------------
// Code
//----------------------------------------------------------
void rx_DrawArea (
		IplImage* ipWkImg,   // 描画するIplImage
		CvSeq*    stContour, // 輪郭データ
		int       iLevel,    // 輪郭階層
		int	      iThresh    // この数値より小さい面はゴミと認識して黒で塗る
		) {
		
	/*******************************************************
	** Initialize 
	*******************************************************/
	CvScalar ScContourColor;	// 輪郭を描画する色の設定
	CvPoint* stCoord;

	int	i;
	double dArea;

	/*******************************************************
	** Process 
	*******************************************************/
	//------------------------------------------------------
	//  ノイズ閾値以下の面積は、背景で塗りつぶす
	//  各種輪郭の特徴量の取得
	//------------------------------------------------------
	dArea = fabs(cvContourArea(stContour, CV_WHOLE_SEQ, 0));
	
	//輪郭の描画 (塗りつぶし)
	if ((int)dArea < iThresh) {  // noise
		//ScContourColor = CV_RGB(0, 0, 0);
		//cvDrawContours(ipWkImg, stContour, ScContourColor, ScContourColor, 0, CV_FILLED, 8, cvPoint(0,0));
	} else {
		for (i = 0; i < stContour->total; i++){
			stCoord = CV_GET_SEQ_ELEM (CvPoint, stContour, i);
			iGravityX[iCntArea] += stCoord->x;
			iGravityY[iCntArea] += stCoord->y;
		}
		iGravityX[iCntArea] = iGravityX[iCntArea] / i;
		iGravityY[iCntArea] = iGravityY[iCntArea] / i;
		iArea[iCntArea] = (int)dArea;
		iAreaCount[iCntArea] = stContour->total;
		iCntArea++;
		
		ScContourColor = CV_RGB( 255, 0, 0 );
		cvDrawContours(ipWkImg, stContour, ScContourColor, ScContourColor, 0, CV_FILLED, 8, cvPoint(0,0));
	}

	//------------------------------------------------------
	// Next Process
	//------------------------------------------------------
	if (stContour->h_next != NULL){
		// 次の輪郭データ処理
		rx_DrawArea(ipWkImg, stContour->h_next, iLevel, iThresh);
	}
	if (stContour->v_next != NULL){
		// 子の輪郭データ処理
		rx_DrawAreaChild(ipWkImg, stContour->v_next, iLevel + 1, iThresh);
	}
}

/***********************************************************
** 子の輪郭描画
************************************************************/
void rx_DrawAreaChild(
		IplImage* ipWkImg,   // 描画するIplImage
		CvSeq*    stContour, // 輪郭データ
		int       iLevel,    // 輪郭階層
		int	      iThresh    // 塗りつぶす大きさの閾値
		){

	/*******************************************************
	** Initialize 
	*******************************************************/
	CvScalar ScContourColor;	// 輪郭を描画する色の設定

	/*******************************************************
	** Process 
	*******************************************************/
	//------------------------------------------------------
	// 子供の輪郭抽出結果は赤で塗りつぶす
	//------------------------------------------------------
	ScContourColor = CV_RGB( 255, 0, 0 );
	cvDrawContours(ipWkImg, stContour, ScContourColor, ScContourColor, 0, CV_FILLED, 8, cvPoint(0,0));
	
	//------------------------------------------------------
	// Next Process
	//------------------------------------------------------
	if (stContour->h_next != NULL){
		// 次の輪郭データ処理
		rx_DrawArea(ipWkImg, stContour->h_next, iLevel, iThresh);
	}
	if (stContour->v_next != NULL){
		// 子の輪郭データ処理
		rx_DrawAreaChild(ipWkImg, stContour->v_next, iLevel + 1, iThresh);
	}
}

/***********************************************************
** ラベリング処理
************************************************************/
int cv_Labelling(
		IplImage* ipSrcImg,      // 入力画像
		IplImage* ipDstImg,      // 出力画像
		int	      iThresh        // 塗りつぶす大きさの閾値
		){
	
	int	i;
	int iNumFindConter;
	
	CvSeq*          stContour = NULL;
	CvMemStorage*   memStorage;
	IplImage*       ipGrayImg;
	IplImage*       ipBinImg;
	
	/*******************************************************
	** Initialize 
	*******************************************************/
	if (ipSrcImg == NULL) {
		return -1;
	}
	for (i = 0; i < MAX_GRAVITY; i++) {
		iGravityX[i] = 0;
		iGravityY[i] = 0;
	}
	iCntArea = 0;

	//------------------------------------------------------
	// Image resource create
	//------------------------------------------------------
	memStorage = cvCreateMemStorage (0);
	ipGrayImg  = cvCreateImage(cvGetSize(ipSrcImg), IPL_DEPTH_8U, 1);
	ipBinImg   = cvCreateImage(cvGetSize(ipSrcImg), IPL_DEPTH_8U, 1);
	
	/*******************************************************
	** Process 
	*******************************************************/
	// BGRからグレースケール変換する
	cvCvtColor(ipSrcImg, ipGrayImg, CV_BGR2GRAY);
	
	// 二値化
	cvThreshold (ipGrayImg, ipBinImg, 0, 255, CV_THRESH_BINARY | CV_THRESH_OTSU);
	
	//------------------------------------------------------
	// cvFindContours
	//------------------------------------------------------
	// 輪郭の抽出
	iNumFindConter = cvFindContours(ipBinImg, memStorage, &stContour, sizeof(CvContour),
		CV_RETR_EXTERNAL, CV_CHAIN_APPROX_NONE, cvPoint(0, 0));
	
	//------------------------------------------------------
	// 重心設定、ノイズ処理、描画処理
	//------------------------------------------------------
	if (0 < iNumFindConter){
		// 黒く塗りつぶす
		cvSetZero(ipDstImg);
		// 抽出したデータに対する再帰処理
		rx_DrawArea(ipDstImg, stContour, 1, iThresh);
	}
	
	/*******************************************************
	** Terminate 
	*******************************************************/
	//Release
	cvReleaseMemStorage(&memStorage);
	cvReleaseImage(&ipGrayImg);
	cvReleaseImage(&ipBinImg);
	
	return iCntArea;
}
// EOF