/***********************************************************
Common software for servo function.
Author : Naoki Kabaya
create : 01/10/2013 version 1.0
modify :
***********************************************************/
//----------------------------------------------------------
// Includes
//----------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <memory.h>
#include <unistd.h>   // Linux
#include <termios.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

// OpenCV
#include <cv.h>
#include <highgui.h>
#include <cxcore.h>

//----------------------------------------------------------
// Defines
//----------------------------------------------------------
#define	SIZE_BUFFER	128
#define	VALUE_THRESHHOLD	40
#define	MINI_AREA	50
#define MAX_GRAVITY	512
#define PI	3.141592
#define ROBTO_END	"/tmp/robot.end"

// WRC103LV
#define _POSIX_SOURCE  1	   /* POSIX comliant source (POSIX)*/
#define BAUDRATE       B115200
#define MODEMDEVICE    "/dev/ttyACM0"
#define BUFFSIZE       256
#define FALSE          0
#define TRUE           1
#define WRC_STRAIGHT		0x77
#define WRC_STOP		0x73
#define WRC_TURNRIGHT	0x64
#define WRC_TURNLEFT	0x61

//getValue
#define SIZE_BUF64	64
#define	FIND		1
#define FIND_NO		0

#define	PR(A,B)	printf("--->%s.(%s)\n", A, B)
#define PRDI(A,B) printf("DEBUG--->%s.(%d)\n", A, B)
#define PRDS(A,B) printf("DEBUG--->%s.(%s)\n", A, B)

//----------------------------------------------------------
// Globals
//----------------------------------------------------------
extern int iGravityX[MAX_GRAVITY];
extern int iGravityY[MAX_GRAVITY];
extern int iArea[MAX_GRAVITY];
extern int iAreaCount[MAX_GRAVITY];

// WRC103LV
static int fd = -1;
struct termios oldtio;      /* save current serial port settings */
unsigned char *receive_buffer; 

//----------------------------------------------------------
// Prototype
//----------------------------------------------------------
int main(int argc, char* argv[]);
void cv_HsvExtraction(IplImage* src_img, IplImage* dst_img,
		int iCh1LowerValue, int iCh1UpperValue,int iCh2LowerValue, int iCh2UpperValue,
		int iCh3LowerValue, int iCh3UpperValue);
int cv_Labelling(IplImage* src_img, IplImage* dst_img, int iThresh);
int getValue(char *szFile, char *szKeyword, char *szValue);

// WRC103LV
void serial_init(void);
void serial_close(void);
int serial_write(unsigned char *buf);

/***********************************************************
** controlファイルで設定された情報に基づき処理をする
** 指定された色相画素の抽出
** 認識結果は、赤い領域として表示する
** ノイズ処理は、抽出した領域の面積値
***********************************************************/
int main(int argc, char* argv[]){
	
	char name[SIZE_BUFFER], szValue[64];
	int	iTargetX1, iTargetY1;
	int	iCalibration = 0;
	int iCh1LowerValue, iCh1UpperValue, iCh2LowerValue, iCh2UpperValue, iCh3LowerValue, iCh3UpperValue;
	int	iThresh;
	int	iCamNo, iCamSkip, iW, iH;
	int	iMvXStraight, iMxMagnification;
	int	iMvTurnUnitMsec, iMvGoStraightUnitMsec;
	int	iRet, i, k;
	double dCamW, dCamH;
	
	FILE 		*chkfl;
	time_t 		timer;
    struct tm 	*local;
	CvPoint 	p1, p2;
	
	// WRC103LV
	unsigned char	buf;
	
	// OpenCV font
    CvFont 	dfont;
    float 	hscale;
    float 	vscale;
    float 	italicscale;
    int  	thickness;
    char 	szText[255];
	
	/*******************************************************
	** プログラムの初期化処理
	*******************************************************/
	if (argc != 3) {
		PR("Usage: ", "robot controlFile output-dir");
		return -1;
	}

	//------------------------------------------------------
	// (1) 環境設定情報の読込み
	// iCamNo, iW, iH, iCamSkip,
	// iCh1LowerValue, iCh1UpperValue, iCh2LowerValue, iCh2UpperValue, 
	// iCh3LowerValue, iCh3UpperValue, iTresh, iMvXStraight
	//------------------------------------------------------
	if ((iRet = getValue(argv[1], "WEBCAM_DEVICE", szValue)) != 0) {
		printf("error can't get WEBCAM_DEVICE\n");
		return -1;
	} else {
		iCamNo = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "WEBCAM_WIDTH", szValue)) != 0) {
		printf("error can't get WEBCAM_WIDTH\n");
		return -1;
	} else {
		iW = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "WEBCAM_HIEGHT", szValue)) != 0) {
		printf("error can't get WEBCAM_HIEGHT\n");
		return -1;
	} else {
		iH = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "WEBCAM_SKIPFRAME", szValue)) != 0) {
		printf("error can't get WEBCAM_SKIPFRAME\n");
		return -1;
	} else {
		iCamSkip = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "HUE_LOWERVALUE", szValue)) != 0) {
		printf("error can't get HUE_LOWERVALUE\n");
		return -1;
	} else {
		iCh1LowerValue = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "HUE_UPPERVALUE", szValue)) != 0) {
		printf("error can't get HUE_UPPERVALUE\n");
		return -1;
	} else {
		iCh1UpperValue = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "SATURATION_LOWERVALUE", szValue)) != 0) {
		printf("error can't get SATURATION_LOWERVALUE\n");
		return -1;
	} else {
		iCh2LowerValue = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "SATURATION_UPPERVALUE", szValue)) != 0) {
		printf("error can't get SATURATION_UPPERVALUE\n");
		return -1;
	} else {
		iCh2UpperValue = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "VALUE_LOWERVALUE", szValue)) != 0) {
		printf("error can't get VALUE_LOWERVALUE\n");
		return -1;
	} else {
		iCh3LowerValue = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "VALUE_UPPERVALUE", szValue)) != 0) {
		printf("error can't get VALUE_UPPERVALUE\n");
		return -1;
	} else {
		iCh3UpperValue = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "TRASH_AREASIZE", szValue)) != 0) {
		printf("error can't get TRASH_AREASIZE\n");
		return -1;
	} else {
		iThresh = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "DISTANCE_XAXIS", szValue)) != 0) {
		printf("error can't get DISTANCE_XAXIS\n");
		return -1;
	} else {
		iMvXStraight = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "MAGNIFICATION_X", szValue)) != 0) {
		printf("error can't get MAGNIFICATION_X\n");
		return -1;
	} else {
		iMxMagnification = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "GOSTRAIGHT_UNIT_MSEC", szValue)) != 0) {
		printf("error can't get GOSTRAIGHT_UNIT_MSEC\n");
		return -1;
	} else {
		iMvGoStraightUnitMsec = atoi(szValue);
	}

	// 読取ったデータを画面に表示する
	printf("Parameter %d-%d-%d-%d-%d-%d-%d-%d-%d-%d-%d-%d-%d-%d-%d\n",
	iCamNo, iW, iH, iCamSkip, iCh1LowerValue, iCh1UpperValue,
	iCh2LowerValue, iCh2UpperValue, iCh3LowerValue, iCh3UpperValue,
	iThresh, iMvXStraight, iMxMagnification,
	iMvTurnUnitMsec, iMvGoStraightUnitMsec);
	
	// 型の変更
	dCamW = (float)iW; // webcam width
	dCamH = (float)iH; // webcam hieght

	//------------------------------------------------------
	// (2) Webcamの初期化
	//------------------------------------------------------
	CvCapture* capture = cvCreateCameraCapture (iCamNo);
	if (capture == NULL) { // ビデオカメラが取得できない時
		fprintf(stderr, "There is not the camera.\n");
		return -1;
	}
	cvSetCaptureProperty(capture, CV_CAP_PROP_FRAME_WIDTH, dCamW);
	cvSetCaptureProperty(capture, CV_CAP_PROP_FRAME_HEIGHT, dCamH);

	//------------------------------------------------------
	// (3) 8bit 3channel画像処理メモリーの確保
	//------------------------------------------------------
	// 画像処理用のメモリー領域を確保する
	IplImage* Src_image = cvCreateImage(cvSize(iW, iH), IPL_DEPTH_8U, 3);
	IplImage* Result_image = cvCreateImage(cvGetSize(Src_image), IPL_DEPTH_8U, 3);
	IplImage* contour_image = cvCreateImage(cvGetSize(Src_image), IPL_DEPTH_8U, 3 );

	// 画像保存時のパラメータを設定する
	int param[]={CV_IMWRITE_JPEG_QUALITY, 100};

	//------------------------------------------------------
	// (4) 表示用フォントの初期化
	//------------------------------------------------------
	hscale      = 1.0f;
	vscale      = 1.0f;
	italicscale = 0.0f;
	thickness    = 1;
	cvInitFont (&dfont, CV_FONT_HERSHEY_PLAIN , hscale, vscale, italicscale, thickness, NULL);

	//------------------------------------------------------
	// (5) WRC103LV オープンと初期化
	//------------------------------------------------------
	fd = open(MODEMDEVICE, O_RDWR | O_NOCTTY);
	if (fd < 0) {
		perror(MODEMDEVICE);
		return (-1);
	}
	serial_init();
	
	// Start Moving straight.
	PRDS("START", "WRC103LV STRAIGHT");
	buf = WRC_STRAIGHT;
	iRet = serial_write(&buf);
	usleep(iMvGoStraightUnitMsec);
	
	/*******************************************************
	** 画像認識処理ループ
	*******************************************************/
	while (1) { 
		//------------------------------------------------------
		// (1) 移動制御の終了確認
		//------------------------------------------------------
		if ((chkfl = fopen(ROBTO_END, "r")) != NULL) {
			fclose(chkfl);
			PRDS("STOP", "WRC103LV STRAIGHT");
			buf = WRC_STOP;
			iRet = serial_write(&buf);
			break;
		}
		
		//------------------------------------------------------
		// (2) 起動時のみ実行するWebcam稼動準備
		//------------------------------------------------------
		if (iCalibration == 0) {
			for (i = 0; i < iCamSkip; i++) {
				Src_image = cvQueryFrame (capture);
				iCalibration = 1;
			}
		} else {
			Src_image = cvQueryFrame (capture);
		}

		//------------------------------------------------------
		// (3) 指定した色相の画素を抽出する
		//------------------------------------------------------
		cv_HsvExtraction(Src_image, Result_image,
			iCh1LowerValue, iCh1UpperValue, iCh2LowerValue, iCh2UpperValue, iCh3LowerValue, iCh3UpperValue);
		
		//------------------------------------------------------
		// (4) 上記で抽出した画素の塊からiThresh以上の領域を抽出する
		//     iThreadは、ノイズとして認識する閾値
		//------------------------------------------------------
		iRet = cv_Labelling(Result_image, contour_image, iThresh);
		if (iRet == 0) {
			continue;
		}

		//------------------------------------------------------
		// (5) ボールとして認識した対象領域の重心、座標を描画する
		//     元画像と、処理済み画像に格子線を描画する
		//     格子線は、中心よりiMvXStraightの間隔で左右上下に描画する
		//------------------------------------------------------
		iTargetX1 = 0;
		iTargetY1 = 0;
		if (0 < iRet) {
			for (i = 0; i < iRet; i++) {
				printf("Gravity No. %d  X:%d  Y:5%d\n", i, iGravityX[i], iGravityY[i]);
				iTargetX1 += iGravityX[i];
				iTargetY1 += iGravityY[i];
				
				//Draw gravities points
				cvCircle (contour_image, cvPoint (iGravityX[i], iGravityY[i]), 10, CV_RGB (255, 255, 255), 2, 8, 0);
				
				sprintf(szText, "Area:%d, Coord:%d.", iArea[i], iAreaCount[i]);
				cvPutText(contour_image, szText, cvPoint(iGravityX[i]-100, iGravityY[i]+25), &dfont, CV_RGB(185, 196, 47));
			}

			// 取得した画像と認識後のイメージデータに格子線を描画する
			p1.x = iW / 2;
			p1.y = 0;
			p2.x = iW / 2;
			p2.y = iH;
			cvLine(Src_image, p1, p2, CV_RGB( 255, 255, 255), 2, CV_AA ,0);
			cvLine(contour_image, p1, p2, CV_RGB( 255, 255, 255), 2, CV_AA ,0);
			p1.x = 0;
			p1.y = iH / 2;
			p2.x = iW;
			p2.y = iH / 2;
			cvLine(Src_image, p1, p2, CV_RGB( 255, 255, 255), 2, CV_AA ,0);
			cvLine(contour_image, p1, p2, CV_RGB( 255, 255, 255), 2, CV_AA ,0);
			
			k = ((iW/2) % iMvXStraight);
			for (i = 1; i < (k+1); i++) {
				p1.x = iW / 2 + iMvXStraight * i;
				p1.y = 0;
				p2.x = iW / 2 + iMvXStraight * i;
				p2.y = iH;
				cvLine(Src_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				cvLine(contour_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				p1.x = iW / 2 - (iMvXStraight * i);
				p1.y = 0;
				p2.x = iW / 2 - (iMvXStraight * i);
				p2.y = iH;
				cvLine(Src_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				cvLine(contour_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
			}
			k = ((iH/2) % iMvXStraight);
			for (i = 1; i < (k+1); i++) {
				p1.y = iH / 2 + iMvXStraight * i;
				p1.x = 0;
				p2.y = iH / 2 + iMvXStraight * i;
				p2.x = iW;
				cvLine(Src_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				cvLine(contour_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				p1.y = iH / 2 - (iMvXStraight * i);
				p1.x = 0;
				p2.y = iH / 2 - (iMvXStraight * i);
				p2.x = iW;
				cvLine(Src_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				cvLine(contour_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
			}
		}
		
		//------------------------------------------------------
		// (5) 画像を保存する
		//------------------------------------------------------
		// set file name
		memset((void *)name , (int)'\0', sizeof(name));
		timer = time(NULL);
		local = localtime(&timer);
		
		sprintf(name, "%s/%4d%02d%02d%02d%02d%02d.jpg",
			argv[2], local->tm_year+1900, local->tm_mon+1, 
			local->tm_mday, local->tm_hour, local->tm_min, local->tm_sec);
		cvSaveImage(name, contour_image, param);
		
		sprintf(name, "%s/%4d%02d%02d%02d%02d%02d-src.jpg",
			argv[2], local->tm_year+1900, local->tm_mon+1, 
			local->tm_mday, local->tm_hour, local->tm_min, local->tm_sec);
		cvSaveImage(name, Src_image, param);

		//------------------------------------------------------
		// (6) WRC103LV 移動方向の決定
		//------------------------------------------------------
		// TRASH_AREASIZE以上のノイズが有った場合、急激な回転をしないように
		// 平均値を使って、中心との変位X値を求める
		iTargetX1 = (iTargetX1 / iRet) - (iW / 2);
		if (iTargetX1 == 0) {
			continue;
		}

		// 変位値が、直進の範囲以上であれば方向変更する
		if ((iMvXStraight * iMxMagnification) < abs(iTargetX1)) { // Turn
			if (0 < iTargetX1) {
				buf = WRC_TURNRIGHT; // turn right
			} else {
				buf = WRC_TURNLEFT;	// turn left
			}
			
			// WRC103LVに方向変更を指示する
			iRet = serial_write(&buf);
			if (iRet != 0) {
				printf("WRC103LV ERROR in control moving.\n");
				buf = WRC_STOP;
				iRet = serial_write(&buf);
				break;
			}
		} else { // 直進する
			// WRC103LVに直進を指示する
			buf = WRC_STRAIGHT;
			iRet = serial_write(&buf);
			if (iRet != 0) {
				printf("WRC103LV ERROR in control moving.\n");
				buf = WRC_STOP;
				iRet = serial_write(&buf);
				break;
			}
		} //turn end
		// 一定時間同じ動きをする
		usleep(iMvGoStraightUnitMsec);
	} // end of while 

	/*******************************************************
	** プログラムの終了処理
	*******************************************************/
	// WRC103LV terminate
	buf = WRC_STOP;
	iRet = serial_write(&buf);
	serial_close();
	PRDS("END", "WRC103LV teminate end");
	
	// release maemory
	cvReleaseCapture(&capture);
	cvReleaseImage(&Src_image);
	cvReleaseImage(&Result_image);
	cvReleaseImage(&contour_image);
	
	return 0;
}

/***********************************************************
** controlファイルに設定されている値を、
** KEYWORD指定で、文字列として取出す
************************************************************/
int getValue(char *szFile, char *szKeyword, char *szValue){
	FILE	*infl;
	char	szBuf[SIZE_BUF64];
	char	*cRet, *cWord;
	int		j, iFlag;
	
	if ((infl = fopen(szFile, "r")) ==NULL) {
		return(-1);
	}

	iFlag = FIND_NO;
	while (fgets(szBuf, SIZE_BUF64, infl) != NULL) {
		if (strlen(szBuf) < 2) {
			continue;
		}
		if (szBuf[0] == '#' ||  szBuf[0] == ' ') {
			continue;
		}

		if ((cRet = strstr(szBuf, szKeyword)) != NULL) {
			iFlag = FIND;
			cWord = strtok(szBuf, " ,	");
			cWord = strtok(NULL, " ,	");
			j = (int)strlen(cWord);
			*(cWord + j - 1)= '\0';
			strcpy(szValue, cWord);
			break;
		}
	} //end of while
	fclose(infl);
	if (iFlag == FIND_NO) {
		return -1;
	}
	return 0;
}

/***********************************************************
** WRC103LV serial communication
************************************************************/
void serial_init() {
	struct termios tio;
	memset(&tio, 0, sizeof(tio));

	tcgetattr(fd, &oldtio);
	
	tio.c_iflag     = IGNPAR | ICRNL;
	tio.c_iflag     = 0;
	tio.c_oflag     = 0;
	tio.c_lflag     = ICANON;
	tio.c_lflag     = 0;
	tio.c_cflag     = CS8 | CLOCAL | CREAD;
	tio.c_cc[VTIME] = 10; 
	tio.c_cc[VMIN] = 1; 

	cfsetispeed(&tio, BAUDRATE);
	cfsetospeed(&tio, BAUDRATE);
	tcsetattr(fd, TCSANOW, &tio);
}

void serial_close() {
	/* Restore the old port settings */
	tcsetattr(fd, TCSANOW, &oldtio);
	close(fd);
}

int serial_write(unsigned char *buf) {
	unsigned char	sz[128];
	const char *str;

	// Write WRC103LV
	if (write(fd, buf, 1) != 1) {
		printf("WRC103LV write failed.\n");
		read(fd, (void *)str, strlen(str));
		printf("read serial.(%s)\n", sz);
		return -1;
	}
	return 0;
}
