/***********************************************************
Common software for servo function.
Author : Naoki Kabaya
create : 01/10/2013 version 1.0
modify :
***********************************************************/
//----------------------------------------------------------
// Includes
//----------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <memory.h>
#include <unistd.h>   // Linux
#include <termios.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

// OpenCV
#include <cv.h>
#include <highgui.h>
#include <cxcore.h>

//----------------------------------------------------------
// Defines
//----------------------------------------------------------
#define	SIZE_BUFFER	128
#define	VALUE_THRESHHOLD	40
#define	MINI_AREA	50
#define MAX_GRAVITY	512

//getValue
#define SIZE_BUF64	64
#define	FIND		1
#define FIND_NO		0

#define	PR(A,B)	printf("--->%s.(%s)\n", A, B)
#define PRDI(A,B) printf("DEBUG--->%s.(%d)\n", A, B)
#define PRDS(A,B) printf("DEBUG--->%s.(%s)\n", A, B)

//----------------------------------------------------------
// Globals
//----------------------------------------------------------
extern int iGravityX[MAX_GRAVITY];
extern int iGravityY[MAX_GRAVITY];
extern int iArea[MAX_GRAVITY];
extern int iAreaCount[MAX_GRAVITY];

//----------------------------------------------------------
// Prototype
//----------------------------------------------------------
int main(int argc, char* argv[]);
void cv_HsvExtraction(IplImage* src_img, IplImage* dst_img,
		int iCh1LowerValue, int iCh1UpperValue,int iCh2LowerValue, int iCh2UpperValue,
		int iCh3LowerValue, int iCh3UpperValue);
int cv_Labelling(IplImage* src_img, IplImage* dst_img, int iThresh);
int getValue(char *szFile, char *szKeyword, char *szValue);

/***********************************************************
** controlファイルで設定された情報に基づき処理をする
** 指定された色相画素の抽出
** 認識結果は、赤い領域として表示する
** ノイズ処理は、抽出した領域の面積値
***********************************************************/
int main(int argc, char* argv[]){
	
	char name[SIZE_BUFFER], szValue[64];
	int	iTargetX1, iTargetY1;
	int	iCalibration = 0;
	int iCh1LowerValue, iCh1UpperValue, iCh2LowerValue, iCh2UpperValue, iCh3LowerValue, iCh3UpperValue;
	int	iThresh;
	int	iCamNo, iCamSkip, iW, iH;
	int	iMvXStraight;
	int	iRet, i, k;
	double dCamW, dCamH;

	FILE 		*chkfl;
	time_t 		timer;
    struct tm 	*local;
	CvPoint 	p1, p2;
	
	// OpenCV font
    CvFont 	dfont;
    float 	hscale;
    float 	vscale;
    float 	italicscale;
    int  	thickness;
    char 	szText[255];
	
	/*******************************************************
	** プログラムの初期化処理
	*******************************************************/
	if (argc != 3) {
		PR("Usage: ", "robotEye controlFile output-dir");
		return -1;
	}

	//------------------------------------------------------
	// (1) 環境設定情報の読込み
	// iCamNo, iW, iH, iCamSkip,
	// iCh1LowerValue, iCh1UpperValue, iCh2LowerValue, iCh2UpperValue, 
	// iCh3LowerValue, iCh3UpperValue, iTresh, iMvXStraight
	//------------------------------------------------------
	if ((iRet = getValue(argv[1], "WEBCAM_DEVICE", szValue)) != 0) {
		printf("error can't get WEBCAM_DEVICE\n");
		return -1;
	} else {
		iCamNo = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "WEBCAM_WIDTH", szValue)) != 0) {
		printf("error can't get WEBCAM_WIDTH\n");
		return -1;
	} else {
		iW = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "WEBCAM_HIEGHT", szValue)) != 0) {
		printf("error can't get WEBCAM_HIEGHT\n");
		return -1;
	} else {
		iH = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "WEBCAM_SKIPFRAME", szValue)) != 0) {
		printf("error can't get WEBCAM_SKIPFRAME\n");
		return -1;
	} else {
		iCamSkip = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "HUE_LOWERVALUE", szValue)) != 0) {
		printf("error can't get HUE_LOWERVALUE\n");
		return -1;
	} else {
		iCh1LowerValue = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "HUE_UPPERVALUE", szValue)) != 0) {
		printf("error can't get HUE_UPPERVALUE\n");
		return -1;
	} else {
		iCh1UpperValue = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "SATURATION_LOWERVALUE", szValue)) != 0) {
		printf("error can't get SATURATION_LOWERVALUE\n");
		return -1;
	} else {
		iCh2LowerValue = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "SATURATION_UPPERVALUE", szValue)) != 0) {
		printf("error can't get SATURATION_UPPERVALUE\n");
		return -1;
	} else {
		iCh2UpperValue = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "VALUE_LOWERVALUE", szValue)) != 0) {
		printf("error can't get VALUE_LOWERVALUE\n");
		return -1;
	} else {
		iCh3LowerValue = atoi(szValue);
		//printf("iCh3LowerValue %d:%s\n", iCh3LowerValue, szValue);
	}
	if ((iRet = getValue(argv[1], "VALUE_UPPERVALUE", szValue)) != 0) {
		printf("error can't get VALUE_UPPERVALUE\n");
		return -1;
	} else {
		iCh3UpperValue = atoi(szValue);
		//printf("iCh3UpperValue %d:%s\n", iCh3UpperValue, szValue);
	}
	if ((iRet = getValue(argv[1], "TRASH_AREASIZE", szValue)) != 0) {
		printf("error can't get TRASH_AREASIZE\n");
		return -1;
	} else {
		iThresh = atoi(szValue);
	}
	if ((iRet = getValue(argv[1], "DISTANCE_XAXIS", szValue)) != 0) {
		printf("error can't get DISTANCE_XAXIS\n");
		return -1;
	} else {
		iMvXStraight = atoi(szValue);
	}

	// 読取ったデータを画面に表示する
	printf("Parameter %d-%d-%d-%d-%d-%d-%d-%d-%d-%d-%d-%d\n",
	iCamNo, iW, iH, iCamSkip, iCh1LowerValue, iCh1UpperValue,
	iCh2LowerValue, iCh2UpperValue, iCh3LowerValue, iCh3UpperValue,
	iThresh, iMvXStraight);
	
	// 型の変更
	dCamW = (float)iW; // webcam width
	dCamH = (float)iH; // webcam hieght

	//------------------------------------------------------
	// (2) Webcamの初期化
	//------------------------------------------------------
	CvCapture* capture = cvCreateCameraCapture (iCamNo);
	if (capture == NULL) { // ビデオカメラが取得できない時
		fprintf(stderr, "There is not the camera.\n");
		return -1;
	}
	cvSetCaptureProperty(capture, CV_CAP_PROP_FRAME_WIDTH, dCamW);
	cvSetCaptureProperty(capture, CV_CAP_PROP_FRAME_HEIGHT, dCamH);
	
/* 所有するwebcamの設定情報が必要なときは、コメントを外して、実行して下さい。
	printf("capture width      %f\n", cvGetCaptureProperty(capture, CV_CAP_PROP_FRAME_WIDTH));
	printf("capture height     %f\n", cvGetCaptureProperty(capture, CV_CAP_PROP_FRAME_HEIGHT));
	printf("capture fps        %f\n", cvGetCaptureProperty(capture, CV_CAP_PROP_FPS));
	printf("capture mode       %f\n", cvGetCaptureProperty(capture, CV_CAP_PROP_MODE));
	printf("capture brightness %f\n", cvGetCaptureProperty(capture, CV_CAP_PROP_BRIGHTNESS));
	printf("capture contrast   %f\n", cvGetCaptureProperty(capture, CV_CAP_PROP_CONTRAST));
	printf("capture saturation %f\n", cvGetCaptureProperty(capture, CV_CAP_PROP_SATURATION));
	printf("capture hue        %f\n", cvGetCaptureProperty(capture, CV_CAP_PROP_HUE));
	printf("capture gain       %f\n", cvGetCaptureProperty(capture, CV_CAP_PROP_GAIN));
	printf("capture exposure   %f\n", cvGetCaptureProperty(capture, CV_CAP_PROP_EXPOSURE));
*/
	//------------------------------------------------------
	// (3) 8bit 3channel画像処理メモリーの確保
	//------------------------------------------------------
	// 画像処理用のメモリー領域を確保する
	IplImage* Src_image = cvCreateImage(cvSize(iW, iH), IPL_DEPTH_8U, 3);
	IplImage* Result_image = cvCreateImage(cvGetSize(Src_image), IPL_DEPTH_8U, 3);
	IplImage* contour_image = cvCreateImage(cvGetSize(Src_image), IPL_DEPTH_8U, 3 );

	// 画像保存時のパラメータを設定する
	int param[]={CV_IMWRITE_JPEG_QUALITY, 100};

	//------------------------------------------------------
	// (4) 表示用フォントの初期化
	//------------------------------------------------------
	hscale      = 1.0f;
	vscale      = 1.0f;
	italicscale = 0.0f;
	thickness    = 1;
	cvInitFont (&dfont, CV_FONT_HERSHEY_PLAIN , hscale, vscale, italicscale, thickness, NULL);

	/*******************************************************
	** 画像認識処理ループ
	*******************************************************/
	while (1) { 
		//------------------------------------------------------
		// (1) 移動制御の終了確認
		//------------------------------------------------------
		if ((chkfl = fopen("/tmp/robotEye.end", "r")) != NULL) {
			fclose(chkfl);
			break;
		}
		
		//------------------------------------------------------
		// (2) 起動時のみ実行するWebcam稼動準備
		//------------------------------------------------------
		if (iCalibration == 0) {
			for (i = 0; i < iCamSkip; i++) {
				Src_image = cvQueryFrame (capture);
				iCalibration = 1;
			}
		} else {
			Src_image = cvQueryFrame (capture);
		}

		//------------------------------------------------------
		// (3) 指定した色相の画素を抽出する
		//------------------------------------------------------
		cv_HsvExtraction(Src_image, Result_image,
			iCh1LowerValue, iCh1UpperValue, iCh2LowerValue, iCh2UpperValue, iCh3LowerValue, iCh3UpperValue);
		
		//------------------------------------------------------
		// (4) 上記で抽出した画素の塊からiThresh以上の領域を抽出する
		//     iThreadは、ノイズとして認識する閾値
		//------------------------------------------------------
		iRet = cv_Labelling(Result_image, contour_image, iThresh);
		if (iRet == 0) {
			continue;
		}

		//------------------------------------------------------
		// (5) ボールとして認識した対象領域の重心、座標を描画する
		//     元画像と、処理済み画像に格子線を描画する
		//     格子線は、中心よりiMvXStraightの間隔で左右上下に描画する
		//------------------------------------------------------
		iTargetX1 = 0;
		iTargetY1 = 0;
		if (0 < iRet) {
			for (i = 0; i < iRet; i++) {
				printf("Gravity No. %d  X:%d  Y:5%d\n", i, iGravityX[i], iGravityY[i]);
				iTargetX1 += iGravityX[i];
				iTargetY1 += iGravityY[i];
				
				//Draw gravities points
				cvCircle (contour_image, cvPoint (iGravityX[i], iGravityY[i]), 10, CV_RGB (255, 255, 255), 2, 8, 0);
				
				sprintf(szText, "Area:%d, Coord:%d.", iArea[i], iAreaCount[i]);
				cvPutText(contour_image, szText, cvPoint(iGravityX[i]-100, iGravityY[i]+25), &dfont, CV_RGB(185, 196, 47));
			}

			// 取得した画像と認識後のイメージデータに格子線を描画する
			p1.x = iW / 2;
			p1.y = 0;
			p2.x = iW / 2;
			p2.y = iH;
			cvLine(Src_image, p1, p2, CV_RGB( 255, 255, 255), 2, CV_AA ,0);
			cvLine(contour_image, p1, p2, CV_RGB( 255, 255, 255), 2, CV_AA ,0);
			cvLine(Result_image, p1, p2, CV_RGB( 255, 255, 255), 2, CV_AA ,0);
			p1.x = 0;
			p1.y = iH / 2;
			p2.x = iW;
			p2.y = iH / 2;
			cvLine(Src_image, p1, p2, CV_RGB( 255, 255, 255), 2, CV_AA ,0);
			cvLine(contour_image, p1, p2, CV_RGB( 255, 255, 255), 2, CV_AA ,0);
			cvLine(Result_image, p1, p2, CV_RGB( 255, 255, 255), 2, CV_AA ,0);
			
			k = ((iW/2) % iMvXStraight);
			for (i = 1; i < (k+1); i++) {
				p1.x = iW / 2 + iMvXStraight * i;
				p1.y = 0;
				p2.x = iW / 2 + iMvXStraight * i;
				p2.y = iH;
				cvLine(Src_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				cvLine(contour_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				cvLine(Result_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				p1.x = iW / 2 - (iMvXStraight * i);
				p1.y = 0;
				p2.x = iW / 2 - (iMvXStraight * i);
				p2.y = iH;
				cvLine(Src_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				cvLine(contour_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				cvLine(Result_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
			}
			k = ((iH/2) % iMvXStraight);
			for (i = 1; i < (k+1); i++) {
				p1.y = iH / 2 + iMvXStraight * i;
				p1.x = 0;
				p2.y = iH / 2 + iMvXStraight * i;
				p2.x = iW;
				cvLine(Src_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				cvLine(contour_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				cvLine(Result_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				p1.y = iH / 2 - (iMvXStraight * i);
				p1.x = 0;
				p2.y = iH / 2 - (iMvXStraight * i);
				p2.x = iW;
				cvLine(Src_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				cvLine(contour_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
				cvLine(Result_image, p1, p2, CV_RGB( 255, 255, 255), 1, CV_AA ,0);
			}
		}
		
		//------------------------------------------------------
		// (5) 画像を保存する
		//------------------------------------------------------
		// set file name
		memset((void *)name , (int)'\0', sizeof(name));
		timer = time(NULL);
		local = localtime(&timer);
		
		sprintf(name, "%s/%4d%02d%02d%02d%02d%02d.jpg",
			argv[2], local->tm_year+1900, local->tm_mon+1, 
			local->tm_mday, local->tm_hour, local->tm_min, local->tm_sec);
		cvSaveImage(name, contour_image, param);
		
		sprintf(name, "%s/%4d%02d%02d%02d%02d%02d-src.jpg",
			argv[2], local->tm_year+1900, local->tm_mon+1, 
			local->tm_mday, local->tm_hour, local->tm_min, local->tm_sec);
		cvSaveImage(name, Src_image, param);

		sprintf(name, "%s/%4d%02d%02d%02d%02d%02d-hsv.jpg",
			argv[2], local->tm_year+1900, local->tm_mon+1, 
			local->tm_mday, local->tm_hour, local->tm_min, local->tm_sec);
		cvSaveImage(name, Result_image, param);
	} // end of while 

	/*******************************************************
	** プログラムの終了処理
	*******************************************************/
	// release maemory
	cvReleaseCapture(&capture);
	cvReleaseImage( &Src_image );
	cvReleaseImage( &Result_image );
	cvReleaseImage( &contour_image );
	
	return 0;
}

/***********************************************************
** controlファイルに設定されている値を、
** KEYWORD指定で、文字列として取出す
************************************************************/
int getValue(char *szFile, char *szKeyword, char *szValue){
	FILE	*infl;
	char	szBuf[SIZE_BUF64];
	char	*cRet, *cWord;
	int		j, iFlag;
	
	if ((infl = fopen(szFile, "r")) ==NULL) {
		return(-1);
	}

	iFlag = FIND_NO;
	while (fgets(szBuf, SIZE_BUF64, infl) != NULL) {
		if (strlen(szBuf) < 2) {
			continue;
		}
		if (szBuf[0] == '#' ||  szBuf[0] == ' ') {
			continue;
		}

		if ((cRet = strstr(szBuf, szKeyword)) != NULL) {
			iFlag = FIND;
			cWord = strtok(szBuf, " ,	");
			cWord = strtok(NULL, " ,	");
			j = (int)strlen(cWord);
			*(cWord + j - 1)= '\0';
			strcpy(szValue, cWord);
			break;
		}
	} //end of while
	fclose(infl);
	if (iFlag == FIND_NO) {
		return -1;
	}
	return 0;
}
